/* gdevqfax.c */
/* QFAX Driver for Ghostscript. */

#include "gdevprn.h"
#define GOTUC
#include "qfaxparse.h"
#include "g3encoder.h"

#ifdef __PROTOTYPES__
#	define PROTO_ARGS(X)	X
#else
#	define PROTO_ARGS(X)	()
#endif

#define offsetof(sname,fname) ((long)&((sname *)0)->fname)

/**********************************************************************/
/*
 *	Generic fax output library
 */
/**********************************************************************/

/*************************************************************************
	Redefine the device descriptor
 *************************************************************************/
struct gx_device_qfax_s {
	gx_device_common;
	gx_prn_device_common;
	CODERSTATE *fax;
};
typedef struct gx_device_qfax_s gx_device_qfax;

/* The device descriptor */
#define X_DPI 204
#define Y_DPI 196
#define LINE_SIZE ((X_DPI * 85 / 10 + 7) / 8)	/* bytes per line */

private dev_proc_open_device(qfax_prn_open);
private dev_proc_print_page(qfax_print_page);
private dev_proc_close_device(qfax_prn_close);

gx_device_procs qfax_std_procs =
	prn_procs(qfax_prn_open, gdev_prn_output_page, qfax_prn_close);

gx_device_qfax gs_qfaxhigh_device =
{	prn_device_std_body(
		gx_device_qfax,
		qfax_std_procs,
		"qfaxhigh",
		85,			/* width_10ths, 8.5" */
		110,			/* height_10ths, 11" */
		X_DPI, Y_DPI,
		0,0,0,0,		/* margins */
		1,
		qfax_print_page
	)
};
gx_device_qfax gs_qfaxhigh2_device =
{	prn_device_std_body(
		gx_device_qfax,
		qfax_std_procs,
		"qfaxhigh2",
		85,			/* width_10ths, 8.5" */
		110,			/* height_10ths, 11" */
		X_DPI, Y_DPI,
		0,0,0,0,		/* margins */
		1,
		qfax_print_page
	)
};

gx_device_qfax gs_qfaxlow_device =
{	prn_device_std_body(
		gx_device_qfax,
		qfax_std_procs,
		"qfaxlow",
		85,			/* width_10ths, 8.5" */
		110,			/* height_10ths, 11" */
		X_DPI, Y_DPI/2,
		0,0,0,0,		/* margins */
		1,
		qfax_print_page
	)
};

gx_device_qfax gs_qfaxlow2_device =
{	prn_device_std_body(
		gx_device_qfax,
		qfax_std_procs,
		"qfaxlow2",
		85,			/* width_10ths, 8.5" */
		110,			/* height_10ths, 11" */
		X_DPI, Y_DPI/2,
		0,0,0,0,		/* margins */
		1,
		qfax_print_page
	)
};


/*************************************************************************
	Driver entry points
 *************************************************************************/

static short mpage = 0;

private int
qfax_prn_open(gx_device *pdev)
{
    int		rc;

    rc = gdev_prn_open(pdev);

    return (rc);
}

private int
qfax_print_page(gx_device_printer *pdev, FILE *prn_stream)
{
        gx_device_qfax	*ddev = (gx_device_qfax *) pdev;
	char		data[LINE_SIZE + 4];
	int		lnum;
	int		line_size;
        int             last_line = 0;
 
	/* For some odd reason, the file isn't open until now */

	if ( gdev_prn_file_is_new(pdev) )
        {
            short fcode = 0;
            short lsb = 0;
            FAXDATA qd;
            FaxParams(&qd);
            if(ddev->fax == NULL)
            {
                ddev->fax = FaxInitCoder (1);
                mpage = 0;
            }
            fcode = (strchr(pdev->dname, '2') != NULL);
            ddev->fax->hd = fileno(prn_stream);	
            FaxOpenFile(NULL, (ddev->y_pixels_per_inch > 120.0), fcode, 1728);
            FaxHeader(0, 0, 0, NULL,0);
        }
        else
        {
            ddev->fax->hd = fileno(prn_stream);	
	}        

        mpage++;
        FaxBeginPage();
	memset(data, '\0', sizeof(data));
	
	line_size = gdev_mem_bytes_per_scan_line( (gx_device *) pdev);
	
	for ( lnum = 0; lnum < pdev->height; lnum++ )
	{
	    short j;
    	    gdev_prn_copy_scan_lines(pdev, lnum, (byte *)data, line_size);
            for(j = 0; j < line_size; j++)
            {
                if(*(data + j))
                {
                    last_line = lnum;
                    break;
                }
            }
	}
	for ( lnum = 0; lnum != last_line; lnum++ )
	{
		gdev_prn_copy_scan_lines(pdev, lnum, (byte *)data, line_size);
        	FaxCodeLine ((byte *)data);
	}
        FaxCodeEOP();    
	return 0;
}

private int
qfax_prn_close(gx_device *pdev)
{
	gx_device_qfax	*ddev = (gx_device_qfax *) pdev;
	int		rc;

	if(ddev->fax)
	{
            FaxHeader(mpage, 0, 2, "ghost",0); 
            FaxCloseFile();
            FaxFreeCoder(ddev->fax);
        }
	rc = gdev_prn_close(pdev);
	return(rc);
}
